import argparse 
from kilosort.gui.launch import launcher

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog='Kilosort4',
        description='Spike sorting GUI'
        )
    parser.add_argument(
        '--filename', default=None, type=str,
        help='filename of binary file to process'
        )
    parser.add_argument(
        '--reset', action='store_true',
        help='Clears all cached settings before opening the GUI.'
    )
    parser.add_argument(
        '--skip_load', action='store_true',
        help="Don't load data upon opening GUI, even if auto-load is enabled."
    )

    args = parser.parse_args()
    launcher(args.filename, args.reset, args.skip_load)



import pytest
from pathlib import Path

import numpy as np
import torch

from kilosort import io


data_shape = (1000, 10)

def make_bfile_with_dtype(dtype, fill, directory, device):
    # Create memmap, write to file, close the file again.
    N, C = fill.shape
    path = directory / 'temp_memmap.bin'

    try:
        a = np.memmap(path, mode='w+', shape=(N,C), dtype=dtype)
        a[:] = fill[:]
        a.flush()
        del(a)

        bfile = io.BinaryFiltered(path, C, device=device, dtype=dtype)
        x = bfile[0:100]
        y = bfile.padded_batch_to_torch(0)

        # Memmap should be loaded as the specified dtype, but the Tensor used for
        # computations in BinaryFiltered should always be converted to float32.
        assert bfile.file.dtype == dtype
        assert x.dtype == torch.float32
        assert y.dtype == torch.float32

    finally:
        # Delete memmap file and re-raise exception
        path.unlink()


def test_uint16(torch_device, data_directory):
    d = 'uint16'
    r = np.random.rand(*data_shape)*(2**30)  # scale to nearly max value
    r = r.astype(d)
    make_bfile_with_dtype(d, r, data_directory, torch_device)

def test_int16(torch_device, data_directory):
    d = 'int16'
    r = np.random.rand(*data_shape)*2 - 1    # scale to (-1,1)
    r = (r*(2**14)).astype(d)                # scale to nearly max value
    make_bfile_with_dtype(d, r, data_directory, torch_device)

def test_int32(torch_device, data_directory):
    d = 'int32'
    r = np.random.rand(*data_shape)*2 - 1    # scale to (-1,1)
    r = (r*(2**30)).astype(d)                # scale to nearly max value
    make_bfile_with_dtype(d, r, data_directory, torch_device)

def test_float32(torch_device, data_directory):
    d = 'float32'
    r = np.random.rand(*data_shape)*2 - 1    # scale to (-1,1)
    r = (r*(2**15)).astype(d)                # scale up, but keep precision
    make_bfile_with_dtype(d, r, data_directory, torch_device)

def test_unsupported(torch_device, data_directory):
    for d in ['uint8', 'int8', 'float64']:
        # May or may not cause an error, but should always warn user that these
        # dtypes are not specifically supported and might cause unexpected
        # behavior.
        with pytest.warns(RuntimeWarning):
            r = (np.random.rand(*data_shape)*(2**6)).astype(d)
            try:
                make_bfile_with_dtype(d, r, data_directory, torch_device)
            except:
                # Unsupported dtypes might cause unexpected errors, that's okay
                # as long as the warning is raised first.
                continue



import os
import gc
import logging
import warnings
logger = logging.getLogger(__name__)

from torch.nn.functional import max_pool2d, avg_pool2d, conv1d, max_pool1d
import numpy as np
import torch
from sklearn.cluster import KMeans
from sklearn.decomposition import TruncatedSVD
from tqdm import tqdm

from kilosort.utils import template_path, log_performance


def my_max2d(X, dt):
    Xmax = max_pool2d(
        X.unsqueeze(0), [2*dt[0]+1, 2*dt[1]+1],
        stride=[1,1], padding=[dt[0],dt[1]]
        )    
    return Xmax[0]

def my_sum2d(X, dt):
    Xsum = avg_pool2d(
        X.unsqueeze(0), [2*dt[0]+1, 2*dt[1]+1],
        stride=[1,1], padding=[dt[0],dt[1]]
        )    
    Xsum *= (2*dt[0]+1) * (2*dt[1]+1)
    return Xsum[0]

def extract_snippets(X, nt, twav_min, Th_single_ch, loc_range=[4,5],
                     long_range=[6,30], device=torch.device('cuda')):
    Xabs   = X.abs()
    Xmax   = my_max2d(Xabs, loc_range)
    ispeak = torch.logical_and(Xmax==Xabs, Xabs > Th_single_ch).float()

    ispeak_sum  = my_sum2d(ispeak, long_range)
    is_peak_iso = ((ispeak_sum==1) * (ispeak==1))

    is_peak_iso[:, :nt] = 0
    is_peak_iso[:, -nt:] = 0

    xy = is_peak_iso.nonzero()

    clips = X[xy[:,:1], xy[:,1:2] - twav_min + torch.arange(nt, device=device)]

    return clips

def extract_wPCA_wTEMP(ops, bfile, nt=61, twav_min=20, Th_single_ch=6, nskip=25,
                       device=torch.device('cuda')):

    clips = np.zeros((500000,nt), 'float32')
    i = 0
    for j in range(0, bfile.n_batches, nskip):
        X = bfile.padded_batch_to_torch(j, ops)
        
        clips_new = extract_snippets(X, nt=nt, twav_min=twav_min,
                                     Th_single_ch=Th_single_ch, device=device)

        nnew = len(clips_new)

        if i+nnew>clips.shape[0]:
            break

        clips[i:i+nnew] = clips_new.cpu().numpy()
        i+= nnew 

    clips = clips[:i]
    clips /= (clips**2).sum(1, keepdims=True)**.5

    model = TruncatedSVD(n_components=ops['settings']['n_pcs']).fit(clips)
    wPCA = torch.from_numpy(model.components_).to(device).float()

    with warnings.catch_warnings():
        msg = 'KMeans is known to have a memory leak on Windows with MKL'
        warnings.filterwarnings("ignore", message=msg)
        # Prevents memory leak for KMeans when using MKL on Windows
        nthread = os.environ.get('OMP_NUM_THREADS')
        new_nthread = 7
        if nthread is not None:
            new_nthread = min(int(nthread), new_nthread)
        os.environ['OMP_NUM_THREADS'] = str(new_nthread)
        model = KMeans(n_clusters=ops['settings']['n_templates'], n_init = 10).fit(clips)
        wTEMP = torch.from_numpy(model.cluster_centers_).to(device).float()
        wTEMP = wTEMP / (wTEMP**2).sum(1).unsqueeze(1)**.5
        if nthread is not None:
            os.environ['OMP_NUM_THREADS'] = nthread
        else:
            os.environ.pop('OMP_NUM_THREADS')

    return wPCA, wTEMP

def get_waves(ops, device=torch.device('cuda')):
    dd = np.load(template_path())
    wTEMP = torch.from_numpy(dd['wTEMP']).to(device)
    wPCA = torch.from_numpy(dd['wPCA']).to(device)
    return wPCA, wTEMP

def template_centers(ops):
    shank_idx = ops['kcoords']
    xc = ops['xc']
    yc = ops['yc']
    dmin = ops['settings']['dmin']
    if dmin is None:
        # Try to determine a good value automatically based on contact positions.
        y_uniq = np.unique(yc)
        if y_uniq.size == 1:
            dmin = 1
        else:
            dmin = np.median(np.diff(np.unique(y_uniq)))
    ops['dmin'] = dmin
    ops['dminx'] = dminx = ops['settings']['dminx']

    # Iteratively determine template placement for each shank separately.
    yup = np.array([])
    xup = np.array([])
    for i in np.unique(shank_idx):
        xc_i = xc[shank_idx == i]
        yc_i = yc[shank_idx == i]
        xmin, xmax, ymin, ymax = xc_i.min(), xc_i.max(), yc_i.min(), yc_i.max()

        yup = np.concatenate([yup, np.arange(ymin, ymax+.00001, dmin/2)])
        nx = np.round((xmax - xmin) / (dminx/2)) + 1
        xup = np.concatenate([xup, np.linspace(xmin, xmax, int(nx))])

    ops['yup'] = np.unique(yup)
    ops['xup'] = np.unique(xup)

    return ops


def template_match(X, ops, iC, iC2, weigh, device=torch.device('cuda')):
    nt = ops['nt']
    nt0 = ops['settings']['nt0min']
    nk = ops['settings']['n_templates']
    NT = X.shape[-1]
    Nfilt = iC.shape[1]
    niter = 40
    nb = (NT-1)//niter+1

    W = ops['wTEMP'].unsqueeze(1)
    B = conv1d(X.unsqueeze(1), W, padding=nt//2)
    As    = torch.zeros((Nfilt, NT), device=device)
    Amaxs = torch.zeros((Nfilt, NT), device=device)
    imaxs = torch.zeros((Nfilt, NT), dtype = torch.int64, device=device)
    ti = torch.arange(Nfilt, device = device)
    tj = torch.arange(nb, device = device)

    for t in range(niter):
        A = torch.einsum('ijk, jklm-> iklm', weigh, B[iC,:, nb*t:nb*(t+1)])        
        A = A.transpose(1,2)
        A = A.reshape(-1, Nfilt, A.shape[-1])
        
        #Aa, imax = torch.max(A, 0) 
        Aa, imax = torch.max(A.abs(), 0)
        imax = (1+imax) * A[imax, ti.unsqueeze(-1), tj[:A.shape[-1]]].sign()

        As[:, nb*t:nb*(t+1)] = Aa
        imaxs[:, nb*t:nb*(t+1)] = imax
        Amax = torch.max(Aa[iC2], 0)[0]
        Amaxs[:, nb*t:nb*(t+1)] = Amax

    Amaxs[:,:nt] = 0
    Amaxs[:,-nt:] = 0
    Amaxs  = max_pool1d(Amaxs.unsqueeze(0), (2*nt0+1), stride = 1, padding = nt0).squeeze(0)
    xy = torch.logical_and(Amaxs==As, As > ops['Th_universal']).nonzero()
    imax = imaxs[xy[:,0], xy[:,1]]
    amp = As[xy[:,0], xy[:,1]]

    ssign = imax.sign()
    imax = imax.abs()-1
    adist = B[iC[:, xy[:,0]], imax%nk, xy[:,1]] * ssign

    #adist = B[iC[:, xy[:,0]], imax%nk, xy[:,1]] 
    
    #xy[:,1] -= nt
    return xy, imax, amp, adist


def nearest_chans(ys, yc, xs, xc, nC, device=torch.device('cuda')):
    ds = (ys - yc[:,np.newaxis])**2 + (xs - xc[:,np.newaxis])**2
    iC = np.argsort(ds, 0)[:nC]
    iC = torch.from_numpy(iC).to(device)
    ds = np.sort(ds, 0)[:nC]

    return iC, ds


def yweighted(yc, iC, adist, xy, device=torch.device('cuda')):    

    yy = torch.from_numpy(yc).to(device)[iC]
    cF0 = torch.nn.functional.relu(adist)
    cF0 = cF0/cF0.sum(0)

    yct = (cF0 * yy[:,xy[:,0]]).sum(0)
    return yct

def run(ops, bfile, device=torch.device('cuda'), progress_bar=None,
        clear_cache=False, verbose=False):        
    sig = ops['settings']['min_template_size']
    nsizes = ops['settings']['template_sizes']
    nb = ops['Nbatches']

    if ops['settings']['templates_from_data']:
        logger.info('Re-computing universal templates from data.')
        # Determine templates and PC features from data.
        ops['wPCA'], ops['wTEMP'] = extract_wPCA_wTEMP(
            ops, bfile, nt=ops['nt'], twav_min=ops['nt0min'], 
            Th_single_ch=ops['settings']['Th_single_ch'], nskip=25,
            device=device
            )
    else:
        logger.info('Using built-in universal templates.')
        # Use pre-computed templates.
        ops['wPCA'], ops['wTEMP'] = get_waves(ops, device=device)

    ops = template_centers(ops)
    [ys, xs] = np.meshgrid(ops['yup'], ops['xup'])
    ys, xs = ys.flatten(), xs.flatten()
    logger.info(f'Number of universal templates: {ys.size}')
    xc, yc = ops['xc'], ops['yc']

    nC = ops['settings']['nearest_chans']
    nC2 = ops['settings']['nearest_templates']
    iC, ds = nearest_chans(ys, yc, xs, xc, nC, device=device)

    # Don't use templates that are too far away from nearest channel
    # (use square of max distance since ds are squared distances)
    igood = ds[0,:] <= ops['max_channel_distance']**2
    iC = iC[:,igood]
    ds = ds[:,igood]
    ys = ys[igood]
    xs = xs[igood]
    ops['ycup'], ops['xcup'] = ys, xs

    iC2, _ = nearest_chans(ys, ys, xs, xs, nC2, device=device)

    ds_torch = torch.from_numpy(ds).to(device).float()
    template_sizes = sig * (1+torch.arange(nsizes, device=device))
    weigh = torch.exp(-ds_torch.unsqueeze(-1) / template_sizes**2)
    weigh = torch.permute(weigh, (2, 0, 1)).contiguous()
    weigh = weigh / (weigh**2).sum(1).unsqueeze(1)**.5

    st = np.zeros((10**6, 6), 'float64')
    tF = np.zeros((10**6, nC , ops['settings']['n_pcs']), 'float32')

    k = 0
    nt = ops['nt']
    tarange = torch.arange(-(nt//2),nt//2+1, device = device)
    logger.info('Detecting spikes...')
    prog = tqdm(np.arange(bfile.n_batches), miniters=200 if progress_bar else None, 
                mininterval=60 if progress_bar else None)
    # repeat performance log after every 10 minutes of data
    log_skip = int(600 / (ops['batch_size'] / ops['fs']))
    try:
        for ibatch in prog:
            if ibatch % log_skip == 0:
                log_performance(logger, 'debug', f'Batch {ibatch} of {nb-1} ({100*(ibatch/nb):.1f}%)')

            X = bfile.padded_batch_to_torch(ibatch, ops)
            xy, imax, amp, adist = template_match(X, ops, iC, iC2, weigh, device=device)
            yct = yweighted(yc, iC, adist, xy, device=device)
            nsp = len(xy)

            if k+nsp>st.shape[0]:
                st = np.concatenate((st, np.zeros_like(st)), 0)
                tF = np.concatenate((tF, np.zeros_like(tF)), 0)

            xsub = X[iC[:,xy[:,:1]], xy[:,1:2] + tarange]
            xfeat = xsub @ ops['wPCA'].T
            tF[k:k+nsp] = xfeat.transpose(0,1).cpu().numpy()

            t_shift = ibatch * bfile.batch_downsampling * (ops['batch_size']/ops['fs'])
            st[k:k+nsp,0] = ((xy[:,1].cpu().numpy()-nt)/ops['fs'] + t_shift)
            st[k:k+nsp,1] = yct.cpu().numpy()
            st[k:k+nsp,2] = amp.cpu().numpy()
            st[k:k+nsp,3] = imax.cpu().numpy()
            st[k:k+nsp,4] = ibatch
            st[k:k+nsp,5] = xy[:,0].cpu().numpy()

            k = k + nsp
            if clear_cache:
                gc.collect()
                torch.cuda.empty_cache()

            if progress_bar is not None:
                progress_bar.emit(int((ibatch+1) / bfile.n_batches * 100))
    except:
        logger.exception(f'Error in spikedetect.run on batch {ibatch}')
        try:
            logger.debug(f'X shape: {X.shape}')
            logger.debug(f'xy shape: {xy.shape}')
        except UnboundLocalError:
            # Error happened before one or both of these was assigned,
            # no need to raise an additional error for this.
            pass
        raise
            
    log_performance(logger, 'debug', f'Batch {ibatch} of {nb-1} ({100*(ibatch/nb):.1f}%)')

    st = st[:k]
    tF = tF[:k]
    ops['iC'] = iC
    ops['iC2'] = iC2
    ops['weigh'] = weigh
    return st, tF, ops



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch


COLOR_CODES = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w']
# This is matplotlib tab10 with gray moved to the last position and all 0.5 alpha
PROBE_PLOT_COLORS = np.array([
        [0.12156863, 0.46666667, 0.70588235, 0.5],
        [1.        , 0.49803922, 0.05490196, 0.5],
        [0.17254902, 0.62745098, 0.17254902, 0.5],
        [0.83921569, 0.15294118, 0.15686275, 0.5],
        [0.58039216, 0.40392157, 0.74117647, 0.5],
        [0.54901961, 0.3372549 , 0.29411765, 0.5],
        [0.89019608, 0.46666667, 0.76078431, 0.5],
        [0.7372549 , 0.74117647, 0.13333333, 0.5],
        [0.09019608, 0.74509804, 0.81176471, 0.5],
        [0.49803922, 0.49803922, 0.49803922, 0.25]
    ])


def plot_drift_amount(ops, results_dir):
    plt.style.use('dark_background')
    fig, ax = plt.subplots(1, 1, figsize=(8,8))
    dshift = ops['dshift']
    settings = ops['settings']

    fs = settings['fs']
    NT = settings['batch_size']
    t = np.arange(dshift.shape[0])*(NT/fs)
    for i in range(dshift.shape[1]):
        color = COLOR_CODES[i % len(COLOR_CODES)]
        ax.plot(t, dshift[:,i], c=color)

    ax.set_xlabel('Time (s)')
    ax.set_ylabel('Depth shift (um)')
    fig.suptitle('Drift amount per probe section, across batches')
    fig.tight_layout()

    save_path = results_dir / 'drift_amount.png'
    fig.savefig(save_path, dpi=300)
    plt.style.use('default')
    plt.close(fig)


def plot_drift_scatter(st0, results_dir):
    fig, ax = plt.subplots(1, 1, figsize=(30,14))

    x = st0[:,0]  # spike time in seconds
    y = st0[:,1]  # depth of spike center in microns
    z = st0[:,2]  # spike amplitude (data)
    z[z < 10] = 10
    z[z > 100] = 100
    colors = np.empty((x.shape[0], 4), dtype=float)

    bin_idx = np.digitize(z, np.logspace(1, 2, 90))
    cm = matplotlib.colormaps['binary']
    for i in np.unique(bin_idx):
        # Take mean of all amplitude values within one bin, map to color
        subset = (bin_idx == i)
        a = z[subset].mean()
        colors[subset] = cm(((a-10)/90))
    
    # Scatter of spike depth over time, with color intensity proportional
    # to log of amplitude.
    ax.scatter(x, y, s=3, c=colors)
    ax.set_xlabel('Time (s)', fontsize=22)
    ax.set_ylabel('Depth (um)', fontsize=22)
    fig.suptitle('Spike amplitude across time and depth', fontsize=30)
    fig.tight_layout()

    save_path = results_dir / 'drift_scatter.png'
    fig.savefig(save_path, dpi=300)
    plt.close(fig)


def plot_diagnostics(Wall0, clu0, ops, results_dir):
    plt.style.use('dark_background')
    fig, axes = plt.subplots(2, 2, figsize=(16,16))
    wPCA = ops['wPCA']
    settings = ops['settings']

    # Top left
    t = np.arange(wPCA.shape[1])/(settings['fs']/1000)
    for i in range(wPCA.shape[0]):
        color = COLOR_CODES[i % len(COLOR_CODES)]
        axes[0][0].plot(t, wPCA[i,:].cpu().numpy(), c=color)
    axes[0][0].set_xlabel('Time (s)')
    axes[0][0].set_title('Temporal Features')

    # Top right
    features = torch.linalg.norm(Wall0, dim=2).cpu().numpy()
    axes[0][1].imshow(features.T, aspect='auto', vmin=0, vmax=25, cmap='binary_r')
    axes[0][1].set_xlabel('Channel Number')
    axes[0][1].set_ylabel('Unit Number')
    axes[0][1].set_title('Spatial Features')

    # Comput spike counts and mean amplitudes
    n_units = int(clu0.max()) + 1
    spike_counts = np.zeros(n_units)
    for i in range(n_units):
        spike_counts[i] = (clu0[clu0 == i]).size
    mean_amp = torch.linalg.norm(Wall0, dim=(1,2)).cpu().numpy()

    # Bottom left
    axes[1][0].plot(mean_amp)
    axes[1][0].set_xlabel('Unit Number')
    axes[1][0].set_ylabel('Amplitude (a.u.)')
    axes[1][0].set_title('Unit Amplitudes')

    # Bottom right
    axes[1][1].scatter(np.log(1 + spike_counts), mean_amp, s=3)
    axes[1][1].set_xlabel('Log(1 + Spike Count)')
    axes[1][1].set_ylabel('Amplitude (a.u.)')
    axes[1][1].set_title('Amplitude vs Spike Count')

    fig.tight_layout()
    save_path = results_dir / 'diagnostics.png'
    fig.savefig(save_path, dpi=300)
    plt.style.use('default')
    plt.close(fig)


def plot_spike_positions(clu, is_refractory, results_dir):
    plt.style.use('dark_background')
    fig, ax = plt.subplots(1, 1, figsize=(30,14))

    # 10 colors in palette, last one is gray for non-frefractory
    clu = clu.copy()
    bad_units = np.unique(clu)[is_refractory == 0]
    bad_idx = np.in1d(clu, bad_units)
    clu = np.mod(clu, 9)
    clu[bad_idx] = 9
    colors = np.empty((clu.shape[0], 4), dtype=float)

    # Map modded cluster ids to colors
    for i in range(10):
        subset = (clu == i)
        rgba = PROBE_PLOT_COLORS[i]
        colors[subset] = rgba

    # Get x, y positions, add to scatterplot
    positions = np.load(results_dir / 'spike_positions.npy')
    xs, ys = positions[:,0], positions[:,1]
    ax.scatter(ys, xs, s=3, c=colors)
    ax.set_xlabel('Depth (um)', fontsize=22)
    ax.set_ylabel('Lateral (um)', fontsize=22)
    fig.suptitle('Spike position across probe, colored by cluster', fontsize=30)
    fig.tight_layout()

    save_path = results_dir / 'spike_positions.png'
    fig.savefig(save_path, dpi=300)
    plt.style.use('default')
    plt.close(fig)



import importlib
from pathlib import Path

import numpy as np
from qtpy import QtCore, QtWidgets
from kilosort.io import (
    RecordingExtractorAsArray, BinaryRWFile, spikeinterface_to_binary, load_probe
    )
from kilosort.gui.logger import setup_logger

logger = setup_logger(__name__)

# TODO: Add others after testing them. Full list here:
#       https://spikeinterface.readthedocs.io/en/latest/modules/extractors.html
_SPIKEINTERFACE_IMPORTS = {
    'NWB': 'read_nwb_recording',
    'Blackrock': 'read_blackrock',
    'Neuralynx': 'read_neuralynx',
    'Openephys': 'read_openephys',
    'Intan': 'read_intan'
}


class DataConversionBox(QtWidgets.QWidget):
    disableInput = QtCore.Signal(bool)
    fileObjectLoaded = QtCore.Signal(bool)

    def __init__(self, gui):
        super().__init__()
        self.gui = gui
        self.setWindowTitle('Convert data (REQUIRES SPIKEINTERFACE)')
        self.filename = None
        self.filetype = None
        self.stream_id = None
        self.stream_name = None
        self.data_dtype = None
        self.file_object = None
        self.conversion_thread = None

        if not self.gui.qt_settings.contains('last_data_location'):
            self.gui.qt_settings.setValue(
                'last_data_location', Path('~').expanduser().as_posix()
                )

        layout = QtWidgets.QGridLayout()

        # Top left: filename and filetype selectors
        self.file_button = QtWidgets.QPushButton('Select file')
        self.file_button.clicked.connect(self.select_file)
        self.or_label = QtWidgets.QLabel('OR')
        self.or_label.setSizePolicy(
            QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Expanding
            )
        self.folder_button = QtWidgets.QPushButton('Select folder')
        self.folder_button.clicked.connect(self.select_folder)
        self.filename_input = QtWidgets.QLineEdit()
        self.filename_input.setReadOnly(True)
        layout.addWidget(self.file_button, 0, 0, 1, 2)
        layout.addWidget(self.or_label, 0, 2, 1, 2)
        layout.addWidget(self.folder_button, 0, 4, 1, 2)
        layout.addWidget(self.filename_input, 1, 0, 1, 6)

        self.filetype_text = QtWidgets.QLabel('File type:')
        self.filetype_selector = QtWidgets.QComboBox()
        self.filetype_selector.addItems([''] + list(_SPIKEINTERFACE_IMPORTS.keys()))
        self.filetype_selector.currentTextChanged.connect(self.select_filetype)
        layout.addWidget(self.filetype_text, 2, 0, 1, 3)
        layout.addWidget(self.filetype_selector, 2, 3, 1, 3)
        if self.gui.qt_settings.contains('filetype'):
            filetype = self.gui.qt_settings.value('filetype')
            if filetype is not None:
                self.filetype_selector.setCurrentText(filetype)

        # Top right: kwargs for spikeinterface loader, dtype
        self.stream_id_text = QtWidgets.QLabel('stream_id (optional)')
        self.stream_id_input = QtWidgets.QLineEdit()
        self.stream_id_input.editingFinished.connect(self.set_stream_id)
        if self.gui.qt_settings.contains('stream_id'):
            stream_id = self.gui.qt_settings.value('stream_id')
            if stream_id is not None:
                self.stream_id_input.setText(stream_id)

        self.stream_name_text = QtWidgets.QLabel('stream_name (optional)')
        self.stream_name_input = QtWidgets.QLineEdit()
        self.stream_name_input.editingFinished.connect(self.set_stream_name)
        if self.gui.qt_settings.contains('stream_name'):
            stream_name = self.gui.qt_settings.value('stream_name')
            if stream_name is not None:
                self.stream_name_input.setText(stream_name)

        self.dtype_text = QtWidgets.QLabel("Data dtype (required):")
        self.dtype_note = QtWidgets.QLabel(
            "NOTE: this is the dtype for the new .bin file.\nIf the existing file"
            " is not the same dtype, data may be altered."
        )
        self.dtype_selector = QtWidgets.QComboBox()
        supported_dtypes = BinaryRWFile.supported_dtypes
        self.dtype_selector.addItems([''] + supported_dtypes)
        self.dtype_selector.currentTextChanged.connect(self.set_dtype)
        if self.gui.qt_settings.contains('converter_dtype'):
            dtype = self.gui.qt_settings.value('converter_dtype')
            if dtype is not None:
                self.dtype_selector.setCurrentText(dtype)
    
        layout.addWidget(self.stream_id_text, 0, 10, 1, 3)
        layout.addWidget(self.stream_id_input, 0, 13, 1, 3)
        layout.addWidget(self.stream_name_text, 1, 10, 1, 3)
        layout.addWidget(self.stream_name_input, 1, 13, 1, 3)
        layout.addWidget(self.dtype_text, 2, 10, 1, 3)
        layout.addWidget(self.dtype_selector, 2, 13, 1, 3)
        layout.addWidget(self.dtype_note, 3, 10, 2, 6)

        # Bottom: convert to binary or load with wrapper
        self.spacer = QtWidgets.QLabel('         ')
        self.wrapper_button = QtWidgets.QPushButton('Load As Wrapper')
        self.wrapper_button.clicked.connect(self.load_as_wrapper)
        self.wrapper_button.setToolTip(
            "'Load as Wrapper' will load the data without converting it, which is much"
            " faster.\n However, sorting will be slower and the results will not "
            "be browsable in Phy.\n We recommend using this option to check that "
            "your data is being loaded correctly."
        )
        self.convert_button = QtWidgets.QPushButton('Convert to Binary')
        self.convert_button.clicked.connect(self.convert_to_binary)
        self.convert_input = QtWidgets.QLineEdit()
        self.convert_input.setPlaceholderText(
            "NOTE: Must include .bin extension in filename."
            )
        self.convert_input.setReadOnly(True)
        self.convert_button.setToolTip(
            "'Convert to Binary' will copy the data to a new .bin file.\n"
            "This process is slow for large recordings, but results will be "
            "browsable in Phy.\n We recommend this option for most use cases.\n"
        )

        layout.addWidget(self.spacer, 5, 0, 2, 6)
        layout.addWidget(self.wrapper_button, 6, 0, 1, 6)
        layout.addWidget(self.convert_button, 7, 0, 1, 6)
        layout.addWidget(self.convert_input, 7, 6, 1, 10)


        # TODO: remove this after bug is fixed
        self.crash_warning = QtWidgets.QLabel(
            "NOTE: we are aware of a bug that causes the GUI to crash after data"
            " is converted.\n If you encounter this issue, the new .bin file"
            " should still have been created,\n so you can open the GUI again "
            "and select it like any other binary file."
        )
        layout.addWidget(self.crash_warning, 8, 0, 3, 16)

        self.setLayout(layout)


    @QtCore.Slot()
    def select_file(self):
        options = QtWidgets.QFileDialog.DontUseNativeDialog
        filename, _ = QtWidgets.QFileDialog.getOpenFileNames(
            parent=self,
            caption="Choose file to load data from...",
            directory=self.gui.qt_settings.value('last_data_location'),
            options=options,
        )
        self.filename = filename
        if not isinstance(self.filename, list):
            self.filename = [self.filename]
        self.filename_input.setText(str(filename))
        data_location = Path(filename[0]).parent.as_posix()
        self.gui.qt_settings.setValue('last_data_location', data_location)

    @QtCore.Slot()
    def select_folder(self):
        options = QtWidgets.QFileDialog.DontUseNativeDialog
        directory = QtWidgets.QFileDialog.getExistingDirectory(
            parent=self,
            caption="Choose directory to load data from...",
            directory=self.gui.qt_settings.value('last_data_location'),
            options=options,
        )
        self.filename = directory
        self.filename_input.setText(directory)
        data_location = Path(directory).as_posix()
        self.gui.qt_settings.setValue('last_data_location', data_location)

    @QtCore.Slot()
    def select_filetype(self):
        self.filetype = self.filetype_selector.currentText()
        self.gui.qt_settings.setValue('filetype', self.filetype)

    @QtCore.Slot()
    def set_stream_id(self):
        stream_id = self.stream_id_input.text()
        if stream_id is None or stream_id == '':
            stream_id = None
        else:
            stream_id = stream_id
        self.stream_id = stream_id
        self.gui.qt_settings.setValue('stream_id', stream_id)

    @QtCore.Slot()
    def set_stream_name(self):
        stream_name = self.stream_name_input.text()
        if stream_name == '':
            stream_name = None
        self.stream_name = stream_name
        self.gui.qt_settings.setValue('stream_name', stream_name)

    @QtCore.Slot()
    def set_dtype(self):
        self.data_dtype = self.dtype_selector.currentText()
        self.gui.qt_settings.setValue('converter_dtype', self.data_dtype)

    def load_cached_wrapper(self):
        logger.info('Loading cached file with external format as wrapper...')
        filename = self.gui.qt_settings.value('data_file_path')
        self.filename = [str(f) for f in filename]
        self.filetype = self.gui.qt_settings.value('filetype')
        self.data_dtype = self.gui.qt_settings.value('converter_dtype')
        self.stream_id = self.gui.qt_settings.value('stream_id')
        self.stream_name = self.gui.qt_settings.value('stream_name')
        self.load_as_wrapper(keep_probe=True)

    @QtCore.Slot()
    def load_as_wrapper(self, keep_probe=False):
        if None in [self.filename, self.filetype, self.data_dtype]:
            logger.exception(
                'File name, file type, and data dtype must be specified.'
            )
            logger.exception(
                f'File name: {self.filename}\nFile type: {self.filetype}\n'
                f'Dtype: {self.data_dtype}'
            )
            return

        self.hide()
        recording = get_recording_extractor(
            self.filename, self.filetype, self.stream_id, self.stream_name
            )
        f = RecordingExtractorAsArray(recording)
        self.file_object = f
        update_settings_box(self.gui.settings_box, f.c, f.fs, self.data_dtype)
        self.fileObjectLoaded.emit(True)

        # The np.dtype() wrapper is here because multiple strings can represent
        # the same type, e.g. '<f4' for 'float32'
        if np.dtype(f.dtype) != np.dtype(self.data_dtype):
            logger.warn(
                f'NOTE: SpikeInterface recording has dtype: {f.dtype},\n'
                f'but dtype: {self.data_dtype} was selected in GUI.'
                )

        # TODO: let user decide whether probe info should be exported, and
        #       set path for saving it.
        from probeinterface import write_prb
        try:
            pg = recording.get_probegroup()
            probe_filename = Path(self.filename[0]).parent / 'probe.prb'
            write_prb(probe_filename, pg)
            add_probe_to_settings(self.gui.settings_box, probe_filename)
        except ValueError:
            print(
                'SpikeInterface recording contains no probe information,\n'
                'could not write .prb file.'
            )
            if not keep_probe:
                self.gui.settings_box.clear_probe_selection()

        self.gui.settings_box.check_load()
        self.gui.qt_settings.setValue('load_as_wrapper', True)

    @QtCore.Slot()
    def convert_to_binary(self):
        # TODO: add option to specify chunksize for conversion
        if None in [self.filename, self.filetype, self.data_dtype]:
            logger.exception(
                'File name, file type, and data dtype must be specified.'
            )
            logger.exception(
                f'File name: {self.filename[0]}\nFile type: {self.filetype}\n'
                f'Dtype: {self.data_dtype}'
            )
            return

        options = QtWidgets.QFileDialog.DontUseNativeDialog
        bin_filename, _ = QtWidgets.QFileDialog.getSaveFileName(
            parent=self,
            caption="Specify a .bin file location to save data to...",
            directory=self.gui.qt_settings.value('last_data_location'),
            filter='*.bin',
            options=options,
        )
        data_location = Path(bin_filename).parent.as_posix()
        self.gui.qt_settings.setValue('last_data_location', data_location)

        self.disableInput.emit(True)
        bin_filename = Path(bin_filename)
        self.hide()
        recording = get_recording_extractor(
            self.filename, self.filetype, self.stream_id, self.stream_name
            )

        self.conversion_thread = ConversionThread(self)
        self.conversion_thread.bin_filename = bin_filename
        self.conversion_thread.recording = recording
        self.conversion_thread.start()
        while self.conversion_thread.isRunning():
            QtWidgets.QApplication.processEvents()
        else:
            self.disableInput.emit(False)

        # Get output of conversion, then delete thread.
        c = self.conversion_thread.c
        fs = self.conversion_thread.fs
        probe_filename = self.conversion_thread.probe_filename
        self.conversion_thread = None

        # Update settings with new data file, probe, etc
        settings = self.gui.settings_box
        update_settings_box(settings, c, fs, self.data_dtype)
        if probe_filename is not None:
            add_probe_to_settings(settings, probe_filename)
        else:
            settings.clear_probe_selection()
        settings.data_file_path_input.setText(bin_filename.as_posix())
        settings.data_file_path_input.editingFinished.emit()
        
        self.gui.settings_box.check_load()


# NOTE: spikeinterface should not be imported anywhere else in this module to
#       avoid an explicit dependency.
def get_recording_extractor(filename, filetype, stream_id=None, stream_name=None):
    '''Load recording by importing an extractor from spikeinterface.'''
    extractor_name = _SPIKEINTERFACE_IMPORTS[filetype]
    import spikeinterface as si
    import spikeinterface.extractors as si_ex
    #extractor_module = importlib.import_module('spikeinterface.extractors')
    if filetype == 'NWB':
        kwargs = {}
    else:
        kwargs = {'stream_id': stream_id, 'stream_name': stream_name}

    if not isinstance(filename, list): filename = [filename]
    extractors = [
        getattr(si_ex, extractor_name)(f, **kwargs)
        for f in filename
        ]
    return si.concatenate_recordings(extractors)


def update_settings_box(settings, n_chans, fs, dtype):
    settings.n_chan_bin_input.setText(str(n_chans))
    settings.n_chan_bin_input.editingFinished.emit()
    settings.fs_input.setText(str(fs))
    settings.fs_input.editingFinished.emit()
    data_idx = settings.dtype_selector.findText(dtype)
    settings.dtype_selector.setCurrentIndex(data_idx)


def add_probe_to_settings(settings, probe_filename):
    settings.probe_layout_selector.setCurrentIndex(0)
    probe = load_probe(probe_filename)
    settings.save_probe_selection(probe, probe_filename.name)
    settings.enable_preview_probe()


class ConversionThread(QtCore.QThread):

    def __init__(self, parent, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent

    def run(self):
        _, _, self.c, _, self.fs, self.probe_filename = spikeinterface_to_binary(
            self.recording, self.bin_filename.parent,
            data_name=self.bin_filename.name,
            dtype=self.parent.data_dtype
        )



import numpy as np


# Format for parameter specification:
# parameter: {
#     'gui_name': text displayed next to edit box in GUI
#     'type': callable datatype for this parameter, like int or float.
#     'min': minimum value allowed (inclusive).
#     'max': maximum value allowed (inclusive).
#     'exclude': list of individual values to exclude from allowed range.
#     'default': default value used by gui and API
#     'step': which step of the pipeline the parameter is used in, from:
#             ['data', 'preprocessing', 'spike detection',
#              'clustering', 'postprocessing']
#     'description': Explanation of parameter's use. Populates parameter help
#                    in GUI.
# }

MAIN_PARAMETERS = {
    # NOTE: n_chan_bin must be specified by user when running through API
    'n_chan_bin': {  
        'gui_name': 'number of channels', 'type': int, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': None, 'step': 'data',
        'description':
            """
            Total number of channels in the binary file, which may be different
            from the number of channels containing ephys data. The value of this
            parameter *must* be specified by the user, or `run_kilosort` will
            raise a ValueError.
            """
    },

    'fs': {
        'gui_name': 'sampling frequency', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 30000, 'step': 'data',
        'description':
            """
            Sampling frequency of probe.
            """
    },

    'batch_size': {
        'gui_name': 'batch size', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 60000, 'step': 'data',
        'description':
            """
            Number of samples included in each batch of data.
            """
    },

    'nblocks': {
        'gui_name': 'nblocks', 'type': int, 'min': 0, 'max': np.inf,
        'exclude': [], 'default': 1, 'step': 'preprocessing',
        'description':
            """
            Number of non-overlapping blocks for drift correction
            (additional nblocks-1 blocks are created in the overlaps).
            """
    },

    'Th_universal': {
        'gui_name': 'Th (universal)', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 9, 'step': 'spike detection',
        'description':
            """
            Spike detection threshold for universal templates.
            Th(1) in previous versions of Kilosort.
            """
    },

    'Th_learned': {
        'gui_name': 'Th (learned)', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 8, 'step': 'spike detection',
        'description':
            """
            Spike detection threshold for learned templates.
            Th(2) in previous versions of Kilosort.
            """
    },

    'tmin': {
        'gui_name': 'tmin', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [], 'default': 0, 'step': 'data',
        'description':
            """
            Time in seconds when data used for sorting should begin. By default,
            begins at 0 seconds.
            """
    },

    'tmax': {
        'gui_name': 'tmax', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': np.inf, 'step': 'data',
        'description':
            """
            Time in seconds when data used for sorting should end. By default,
            ends at the end of the recording.
            """
    },
}


EXTRA_PARAMETERS = {
    ### DATA
    'nt': {
        'gui_name': 'nt', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 61, 'step': 'data',
        'description':
            """
            Number of samples per waveform. Also size of symmetric padding
            for filtering.
            """
    },

    'shift': {
        'gui_name': 'shift', 'type': float, 'min': -np.inf, 'max': np.inf,
        'exclude': [], 'default': None, 'step': 'data',
        'description':
            """
            Scalar shift to apply to data before all other operations. In most
            cases this should be left as None, but may be necessary for float32
            data for example. If needed, `shift` and `scale` should be set such
            that data is roughly in the range -100 to +100.
            
            If set, data will be `data = data*scale + shift`.
            """
    },

    'scale': {
        'gui_name': 'scale', 'type': float, 'min': -np.inf, 'max': np.inf,
        'exclude': [], 'default': None, 'step': 'data',
        'description':
            """
            Scaling factor to apply to data before all other operations. In most
            cases this should be left as None, but may be necessary for float32
            data for example. If needed, `shift` and `scale` should be set such
            that data is roughly in the range -100 to +100.
            
            If set, data will be `data = data*scale + shift`.
            """
    },

    'batch_downsampling': {
        'gui_name': 'batch downsampling', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 1, 'step': 'data',
        'description':
            """
            Number of batches skipped for each batch used for sorting. For example,
            if `batch_downsampling = 10`, then only every 10th batch will be used.
            In general, this should be left as the default (using all batches).
            """
    },

    ### PREPROCESSING
    'artifact_threshold': {
        'gui_name': 'artifact threshold', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [], 'default': np.inf, 'step': 'preprocessing',
        'description':
            """
            If a batch contains absolute values above this number, it will be
            zeroed out under the assumption that a recording artifact is present.
            By default, the threshold is infinite (so that no zeroing occurs).
            """
    },

    'nskip': {
        'gui_name': 'nskip', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 25, 'step': 'preprocessing',
        'description':
            """
            Batch stride for computing whitening matrix.
            """
    },

    'whitening_range': {
        'gui_name': 'whitening range', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 32, 'step': 'preprocessing',
        'description':
            """
            Number of nearby channels used to estimate the whitening matrix.
            """
    },

    'highpass_cutoff': {
        'gui_name': 'highpass cutoff', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [], 'default': 300, 'step': 'preprocessing',
        'description':
            """
            Critical frequency for highpass Butterworth filter applied to data.
            """
    },

    'binning_depth': {
        'gui_name': 'binning_depth', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 5, 'step': 'preprocessing',
        'description':
            """
            For drift correction, vertical bin size in microns used for
            2D histogram.
            """
    },

    'sig_interp': {
        'gui_name': 'sig_interp', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 20, 'step': 'preprocessing',
        'description':
            """
            Approximate spatial smoothness scale in units of microns.
            """
    },

    'drift_smoothing': {
        'gui_name': 'drift smoothing', 'type': list, 'min': None, 'max': None,
        'exclude': [], 'default': [0.5, 0.5, 0.5], 'step': 'preprocessing',
        'description':
            """
            Amount of gaussian smoothing to apply to the spatiotemporal drift
            estimation, for correlation, time (units of registration blocks),
            and y (units of batches) axes. The y smoothing has no effect
            for `nblocks = 1`. Adjusting smoothing for the correlation axis
            is not recommended.
            """
    },


    ### SPIKE DETECTION
    # NOTE: if left as None, will be set to `int(20 * settings['nt']/61)`
    'nt0min': {
        'gui_name': 'nt0min', 'type': int, 'min': 0, 'max': np.inf,
        'exclude': [], 'default': None, 'step': 'spike detection',
        'description': 
            """
            Sample index for aligning waveforms, so that their minimum 
            or maximum value happens here. Defaults to 
            `int(20 * settings['nt']/61)`.
            """
    },

    'dmin': {
        'gui_name': 'dmin', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': None, 'step': 'spike detection',
        'description':
            """
            Vertical spacing of template centers used for spike detection,
            in microns. Determined automatically by default.
            """
    },

    'dminx': {
        'gui_name': 'dminx', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 32, 'step': 'spike detection',
        'description':
            """
            Horizontal spacing of template centers used for spike detection,
            in microns. The default 32um should work well for Neuropixels 1 and
            Neuropixels 2 probes. For other probe geometries, try setting this 
            to the median lateral distance between contacts to start.
            """
    },

    'min_template_size': {
        'gui_name': 'min template size', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 10, 'step': 'spike detection',
        'description':
            """
            Standard deviation of the smallest, spatial envelope Gaussian used
            for universal templates.
            """
    },

    'template_sizes': {
        'gui_name': 'template sizes', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 5, 'step': 'spike detection',
        'description':
            """
            Number of sizes for universal spike templates (multiples of the
            min_template_size).
            """
    },

    'nearest_chans': {
        'gui_name': 'nearest chans', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 10, 'step': 'spike detection',
        'description':
            """
            Number of nearest channels to consider when finding local maxima
            during spike detection.
            """
    },

    'nearest_templates': {
        'gui_name': 'nearest templates', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 100, 'step': 'spike detection',
        'description':
            """
            Number of nearest spike template locations to consider when finding
            local maxima during spike detection.
            """
    },

    'max_channel_distance': {
        'gui_name': 'max channel distance', 'type': float, 'min': 1,
        'max': np.inf, 'exclude': [], 'default': 32, 'step': 'spike detection',
        'description':
            """
            Templates farther away than this from their nearest channel will
            not be used. Also limits distance between compared channels during
            clustering.
            """
    },

    'max_peels': {
        'gui_name': 'max peels', 'type': int, 'min': 1, 'max': 10000, 'exclude': [],
        'default': 100, 'step': 'spike detection',
        'description':
        """
        Number of iterations to do over each batch of data in the matching
        pursuit step. More iterations may detect more overlapping spikes.
        """
    },

    'templates_from_data': {
        'gui_name': 'templates from data', 'type': bool, 'min': None, 'max': None,
        'exclude': [], 'default': True, 'step': 'spike detection',
        'description':
            """
            Indicates whether spike shapes used in universal templates should be 
            estimated from the data or loaded from the predefined templates.
            """
    },

    'n_templates': {
        'gui_name': 'n templates', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 6, 'step': 'spike detection',
        'description':
            """
            Number of single-channel templates to use for the universal
            templates (only used if templates_from_data is True).
            """
    },

    'n_pcs': {
        'gui_name': 'n pcs', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 6, 'step': 'spike detection',
        'description':
            """
            Number of single-channel PCs to use for extracting spike features
            (only used if templates_from_data is True).
            """
    },

    'Th_single_ch': {
        'gui_name': 'Th (single channel)', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 6, 'step': 'spike detection',
        'description':
            """
            For single channel threshold crossings to compute universal-
            templates. In units of whitened data standard deviations. 
            """
    },


    ### CLUSTERING
    'acg_threshold': {
        'gui_name': 'acg threshold', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 0.2, 'step': 'clustering',
        'description':
            """
            Fraction of refractory period violations that are allowed in the ACG 
            compared to baseline; used to assign "good" units. 
            """
    },

    'ccg_threshold': {
        'gui_name': 'ccg threshold', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [0], 'default': 0.25, 'step': 'clustering',
        'description':
            """
            Fraction of refractory period violations that are allowed in the CCG
            compared to baseline; used to perform splits and merges.
            """
    },

    'cluster_neighbors': {
        'gui_name': 'cluster neighbors', 'type': int, 'min': 2, 'max': np.inf,
        'exclude': [], 'default': 10, 'step': 'clustering',
        'description':
            """
            Number of nearest spike neighbors to search for in
            `clustering_qr.neigh_mat` when building the adjacency matrix that
            defines the graph for clustering. Note that changes to this parameter
            will affect resource usage and sorting time.
            """ 
    },

    'cluster_downsampling': {
        'gui_name': 'cluster downsampling', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 1, 'step': 'clustering',
        'description':
            """
            Inverse fraction of spikes used as landmarks during clustering. By
            default, all spikes are used up to a maximum of
            `max_cluster_subset=25000`.

            The old default behavior (version < 4.1.0) is
            equivalent to `max_cluster_subset=None, cluster_downsampling=20`.
            """
    },

    'max_cluster_subset': {
        'gui_name': 'max cluster subset', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [np.inf], 'default': 25000, 'step': 'clustering',
        'description':
            """
            Maximum number of spikes to use when searching for nearest neighbors
            to build the graph used for clustering. Within each clustering center,
            only a subset of spikes is searched with the size determined by
            `cluster_downsampling` and the total number of spikes. This sets
            a maximum on the size of that subset, so that it will not grow without
            bound for very long recordings. Using a very large number of spikes
            is not necessary and causes performance bottlenecks.

            Use `max_cluster_subset = None` if you do not want a limit on
            the subset size. The old default behavior (version < 4.1.0) is
            equivalent to `max_cluster_subset=None, cluster_downsampling=20`.

            Note: In practice, the actual number of spikes used may increase or
            decrease slightly while staying under the maximum. This happens
            because the maximum is set by adjusting `cluster_downsampling` on the
            fly so that it results in a set no larger than the given size.
            """
    },

    'x_centers': {
        'gui_name': 'x centers', 'type': int, 'min': 1,
        'max': np.inf, 'exclude': [], 'default': None, 'step': 'clustering',
        'description':
            """
            Number of x-positions to use when determining center points for
            template groupings. If None, this will be determined automatically
            by finding peaks in channel density. For 2D array type probes, we
            recommend specifying this so that centers are placed every few
            hundred microns.
            """
    },

    'cluster_init_seed': {
        'gui_name': 'cluster init seed', 'type': int, 'min': 1, 'max': np.inf,
        'exclude': [], 'default': 5, 'step': 'clustering',
        'description':
            """
            Random seed for kmeans++ algorithm used to initialize the graph
            for clustering.
            """
    },


    ### POSTPROCESSING
    'duplicate_spike_ms': {
        'gui_name': 'duplicate spike ms', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [], 'default': 0.25, 'step': 'postprocessing',
        'description':
            """
            Time in ms for which subsequent spikes from the same cluster are
            assumed to be artifacts. A value of 0 disables this step.

            NOTE: this was formerly handled by `duplicate_spike_bins`, which has
            been deprecated. The new default of 0.25ms is equivalent to the old
            default of 7 bins for a 30kHz sampling rate.
            """
    },

    'position_limit': {
        'gui_name': 'position limit', 'type': float, 'min': 0, 'max': np.inf,
        'exclude': [], 'default': 100, 'step': 'postprocessing',
        'description':
            """
            Maximum distance (in microns) between channels that can be used
            to estimate spike positions in `postprocessing.compute_spike_positions`.
            This does not affect spike sorting, only how positions are estimated
            after sorting is complete.
            """
    },
}

# Add default values to descriptions
for k, v in MAIN_PARAMETERS.items():
    s = f"""
        Default value: {str(v["default"])}   
        Min, max: ({str(v['min'])}, {str(v['max'])})   
        Type: {v['type'].__name__}
        """
    v['description'] += s
for k, v in EXTRA_PARAMETERS.items():
    s = f"""
        Default value: {str(v["default"])}   
        Min, max: ({str(v['min'])}, {str(v['max'])})   
        Type: {v['type'].__name__}
        """
    v['description'] += s

main_defaults = {k: v['default'] for k, v in MAIN_PARAMETERS.items()}
extra_defaults = {k: v['default'] for k, v in EXTRA_PARAMETERS.items()}
# In the format expected by `run_kilosort`
DEFAULT_SETTINGS = {**main_defaults, **extra_defaults}


def compare_settings(settings):
    """Find settings values that differ from the defaults.
    
    Parameters
    ----------
    settings : dict
        Formatted the same as `DEFAULT_SETTINGS`.
    
    Returns
    -------
    modified_settings : dict
        Formatted as above, but only contains keys with values that differ
        from the defaults.
    extra_keys : list
        List of keys that appear in `settings` but not `DEFAULT_SETTINGS`.
        These keys are *not* included in `modified_settings`.

    """
    modified_settings = {}
    extra_keys = []

    for k, v in settings.items():
        if k in DEFAULT_SETTINGS:
            if v != DEFAULT_SETTINGS[k]:
                modified_settings[k] = v
        else:
            extra_keys.append(k)
    return modified_settings, extra_keys



import os

import torch
import numpy as np
from torch.fft import fft, ifft, fftshift
from scipy.interpolate import interp1d
from tqdm import trange

from kilosort import preprocessing
from kilosort.io import BinaryFiltered


if torch.cuda.is_available():
    dev = torch.device('cuda:0')
else:
    dev = torch.device('cpu')


def get_drift_matrix(ops, dshift):
    """ for a given dshift drift, computes the linear drift matrix for interpolation
    """

    # first, interpolate drifts to every channel
    yblk = ops['yblk']
    finterp = interp1d(yblk, dshift, fill_value="extrapolate", kind = 'linear')
    shifts = finterp(ops['probe']['yc'])

    # compute coordinates of desired interpolation
    xp = np.vstack((ops['probe']['xc'],ops['probe']['yc'])).T
    yp = xp.copy()
    yp[:,1] -= shifts

    xp = torch.from_numpy(xp).to(dev)
    yp = torch.from_numpy(yp).to(dev)

    # run interpolated to obtain a kernel
    Kyx = preprocessing.kernel2D_torch(yp, xp, ops['settings']['sig_interp'])
    
    # multiply with precomputed kernel matrix of original channels
    M = Kyx @ ops['iKxx']

    return M


def avg_wav(bf, Wsub, nn, ops, ibatch, st_i, clu, Nfilt, nt, NT):

    if isinstance(ops['wPCA'], np.ndarray):
        ops['wPCA'] = torch.from_numpy(ops['wPCA']).to(dev)


    X = bf.padded_batch_to_torch(ibatch, ops)

    ix = (st_i - bf.imin - NT*ibatch) * (st_i - bf.imin - NT*(1+ibatch)) < 0
    st_sub = st_i[ix] + nt - bf.imin - NT*ibatch
    st_sub = torch.from_numpy(st_sub).to(dev)

    tiwave = torch.arange(nt, device=dev)
    xsub = X[:, st_sub.unsqueeze(-1) + tiwave] @ ops['wPCA'].T
    nsp = xsub.shape[1]
    M = torch.zeros((Nfilt, nsp), dtype=torch.float, device = dev)
    M[clu[ix], torch.arange(nsp)] = 1
    
    Wsub += torch.einsum('ijk, lj -> lki', xsub, M)
    nn += M.sum(1)

    return Wsub, nn


def clu_ypos(filename, ops, st_i, clu, tmin=0.0, tmax=np.inf):
    Nfilt = clu.max()+1
    Wsub = torch.zeros((Nfilt, ops['n_pcs'], ops['Nchan']), device = dev)
    nn   = torch.zeros((Nfilt, ), device = dev)
    fs = ops['fs']
    n_chan_bin = ops['n_chan_bin']
    nt = ops['nt']
    NT = ops['batch_size']
    bf = BinaryFiltered(filename, n_chan_bin=n_chan_bin, fs=fs, NT=NT, nt=nt,
                        hp_filter=ops['fwav'], whiten_mat=ops['Wrot'],
                        dshift=ops['dshift'], chan_map=ops['chanMap'],
                        tmin=tmin, tmax=tmax)

    # Only use batches within tmin, tmax.
    n_batches = min(int((bf.imax-bf.imin)/NT), ops['Nbatches'])
    for ibatch in range(0, n_batches, 10):
        Wsub, nn = avg_wav(bf, Wsub, nn, ops, ibatch, st_i, clu, Nfilt, nt, NT)

    Wsub = Wsub / nn.unsqueeze(-1).unsqueeze(-1)
    Wsub = Wsub.cpu().numpy()

    ichan = np.argmax((Wsub**2).sum(1), -1)
    yclu = ops['yc'][ichan]

    return yclu, Wsub


def nmatch(ss0, ss, dt=6):
    i = 0
    j = 0
    n0 = 0

    ntmax = len(ss0)
    is_matched  = np.zeros(len(ss), 'bool')
    is_matched0 = np.zeros(len(ss0), 'bool')

    while i<len(ss):
        while j+1<ntmax and ss0[j+1] <= ss[i]+dt:
            j += 1

        if np.abs(ss0[j] - ss[i]) <=dt:
            n0 += 1
            is_matched[i] = 1
            is_matched0[j] = 1

        i+= 1
    return n0, is_matched, is_matched0


def match_neuron(kk, clu, yclu, st_i, clu0, yclu0, st0_i, n_check=20, dt=6):
    ss = st_i[clu==kk]
    if len(ss) == 0:
        raise ValueError(f'No GT spikes matching cluster {kk}')
    isort = np.argsort(np.abs(yclu[kk] - yclu0))
    fmax = 0
    miss = 1
    fpos = 0
    best_ind = isort[0]
    matched_all = -1 * np.ones((n_check,))
    top_inds = -1 * np.ones((n_check,), "int")
    ntest = min(len(isort), n_check)
    top_inds[:ntest] = isort[:ntest]

    no_match = 0
    for j in range(ntest):
        ss0 = st0_i[clu0==isort[j]]
        if len(ss0) == 0:
            no_match += 1
            continue
        
        n0, is_matched, is_matched0 = nmatch(ss0, ss, dt=dt)

        #fmax_new = n0 / (len(ss) + len(ss0) - n0)
        fmax_new = 1 - np.maximum(0, 1 - n0/len(ss)) - np.maximum(0, 1 - n0/len(ss0))
        matched_all[j] = n0

        if fmax_new > fmax:
            miss = np.maximum(0, 1 - n0/len(ss))
            fpos = np.maximum(0, 1 - n0/len(ss0))

            fmax = fmax_new
            best_ind = isort[j]
    if no_match == ntest:
        raise ValueError(f'No matching clusters found for cluster {kk}')

    return fmax, miss, fpos, best_ind, matched_all, top_inds


def compare_recordings(st_gt, clu_gt, yclu_gt, st_new, clu_new, yclu_new):
    NN = len(yclu_gt)

    n_check = 20
    fmax = np.zeros(NN,)
    matched_all = np.zeros((NN, n_check))
    fmiss = np.full(NN, np.nan)
    fpos = np.full(NN, np.nan)
    best_ind = np.zeros(NN, "int")
    top_inds = np.zeros((NN, n_check), "int")

    n_skipped = 0
    for kk in trange(NN):
        try:
            out = match_neuron(kk, clu_gt, yclu_gt, st_gt, clu_new, yclu_new, st_new, n_check=n_check)
            fmax[kk], fmiss[kk], fpos[kk], best_ind[kk], matched_all[kk], top_inds[kk] = out
        except ValueError:
            n_skipped += 1
            continue
    print(f'Num skipped clusters: {n_skipped}')

    return fmax, fmiss, fpos, best_ind, matched_all, top_inds


def load_GT(filename, ops, gt_path, toff=20, nmax=600, tmin=0.0, tmax=np.inf):
    #gt_path = os.path.join(ops['data_folder'] , "sim.imec0.ap_params.npz")
    dd = np.load(gt_path, allow_pickle = True)

    st_gt = dd['st'].astype('int64')
    clu_gt = dd['cl'].astype('int64')

    idx = get_valid_times(st_gt, tmin, tmax, ops['fs'])
    st_gt = st_gt[idx]
    clu_gt = clu_gt[idx]

    ix = clu_gt<nmax
    st_gt  = st_gt[ix]
    clu_gt = clu_gt[ix]

    yclu_gt, Wsub = clu_ypos(filename, ops, st_gt - toff, clu_gt.astype('int64'),
                             tmin=tmin, tmax=tmax)
    mu_gt = (Wsub**2).sum((1,2))**.5

    unq_clu, nsp = np.unique(clu_gt, return_counts = True)
    if np.abs(np.diff(unq_clu) - 1).sum()>1e-5:
        print('error, some ground truth units are missing')

    return st_gt, clu_gt, yclu_gt, mu_gt, Wsub, nsp


def convert_ks_output(ops, st, clu, toff = 20):
    st = st[:,0].astype('int64')        
    yclu, Wsub    = clu_ypos(ops, st-toff, clu)

    return st, clu, yclu, Wsub


def load_phy(filename, fpath, ops, tmin=0.0, tmax=np.inf):
    st_new  = np.load(os.path.join(fpath,  "spike_times.npy")).astype('int64')
    try:
        clu_new = np.load(os.path.join(fpath ,"cluster_times.npy")).astype('int64')
    except:
        clu_new = np.load(os.path.join(fpath ,"spike_clusters.npy")).astype('int64')
    if st_new.ndim==2:
        st_new = st_new[:,0]
    if clu_new.ndim==2:
        clu_new = clu_new[:,0]

    idx = get_valid_times(st_new, tmin, tmax, ops['fs'])
    st_new = st_new[idx]
    clu_new = clu_new[idx]

    yclu_new, Wsub = clu_ypos(filename, ops, st_new - 20, clu_new,
                              tmin=tmin, tmax=tmax)

    return st_new, clu_new, yclu_new, Wsub


def get_valid_times(st, tmin, tmax, fs):
    imin = int(tmin * fs)
    if tmax < np.inf:
        imax = int(tmax * fs)
    else:
        imax = np.inf
    idx = np.logical_and(st >= imin, st < imax)

    return idx



import os, sys, tempfile, shutil, pathlib, psutil
import subprocess
import time
import urllib
import importlib.util
import logging
import pprint
logger = logging.getLogger(__name__)

import numpy as np
from tqdm import tqdm

import torch

# pynvml is an optional dependency for Pytorch cuda, so existing installations
# may or may not have it. It's not worth causing an interruption for this.
if importlib.util.find_spec('pynvml') is not None:
    _NVML_EXISTS = True
else:
    _NVML_EXISTS = False

_DOWNLOADS_DIR_ENV = os.environ.get("KILOSORT_LOCAL_DOWNLOADS_PATH")
_DOWNLOADS_DIR_DEFAULT = pathlib.Path.home().joinpath('.kilosort')
DOWNLOADS_DIR = pathlib.Path(_DOWNLOADS_DIR_ENV) if _DOWNLOADS_DIR_ENV else _DOWNLOADS_DIR_DEFAULT
PROBE_DIR = DOWNLOADS_DIR.joinpath('probes')
PROBE_URLS = {
    # Same as Linear16x1_kilosortChanMap.mat
    'Linear16x1_test.mat': 'https://osf.io/download/67f012cbc56bef203cb25416/',
    # Same as neuropixPhase3B1_kilosortChanMap.mat
    'NeuroPix1_default.mat': 'https://osf.io/download/67f012cc7e1fd38cad82980a/',
    # Same as NP2_kilosortChanMap.mat
    'NeuroPix2_default.mat': 'https://osf.io/download/67f012ce033d25194f829812/'
}


def template_path(basename='wTEMP.npz'):
    """ currently only one set of example templates to use"""
    return cache_template_path(basename)


def cache_template_path(basename):
    cached_file = os.fspath(DOWNLOADS_DIR.joinpath(basename)) 
    if not os.path.exists(cached_file):
        url = 'https://osf.io/download/6807fb5958b763aae139aa60/'
        logger.info('Downloading: "{}" to {}\n'.format(url, cached_file))
        download_url_to_file(url, cached_file, progress=True)
    return cached_file


def download_probes(probe_dir=None):
    if probe_dir is None:
        probe_dir = PROBE_DIR
    probe_dir.mkdir(parents=True, exist_ok=True)
    for probe_name, url in PROBE_URLS.items():
        cached_file = os.fspath(probe_dir.joinpath(probe_name)) 
        if not os.path.exists(cached_file):
            logger.info('Downloading: "{}" to {}\n'.format(url, cached_file))
            try:
                download_url_to_file(url, cached_file, progress=True)
            except urllib.error.HTTPError as e:
                logger.info(f'Unable to download probe {probe_name}, error:')
                logger.info(e)


def retry_download(func, n_tries=5):
    def retry(*args, **kwargs):
        i = 0
        while True:
            try:
                func(*args, **kwargs)
                break
            except urllib.error.HTTPError as e:
                # Try it several times, wait a couple seconds between attempts.
                if i < n_tries:
                    print(f'Download failed, retrying... {i}/{n_tries-1}')
                    i += 1
                    time.sleep(1)
                else:
                    raise e
    
    return retry


@retry_download
def download_url_to_file(url, dst, progress=True):
    r"""Download object at the given URL to a local path.
            Thanks to torch, slightly modified
    Args:
        url (string): URL of the object to download
        dst (string): Full path where object will be saved, e.g. `/tmp/temporary_file`
        progress (bool, optional): whether or not to display a progress bar to stderr
            Default: True
    """
    file_size = None
    import ssl
    ssl._create_default_https_context = ssl._create_unverified_context
    u = urllib.request.urlopen(url)
    meta = u.info()
    if hasattr(meta, 'getheaders'):
        content_length = meta.getheaders("Content-Length")
    else:
        content_length = meta.get_all("Content-Length")
    if content_length is not None and len(content_length) > 0:
        file_size = int(content_length[0])
    # We deliberately save it in a temp file and move it after
    dst = os.path.expanduser(dst)
    dst_dir = os.path.dirname(dst)
    f = tempfile.NamedTemporaryFile(delete=False, dir=dst_dir)
    try:
        with tqdm(total=file_size, disable=not progress,
                  unit='B', unit_scale=True, unit_divisor=1024) as pbar:
            while True:
                buffer = u.read(8192)
                if len(buffer) == 0:
                    break
                f.write(buffer)
                pbar.update(len(buffer))
        f.close()
        shutil.move(f.name, dst)
    finally:
        f.close()
        if os.path.exists(f.name):
            os.remove(f.name)


def get_performance():
    """Get resource usage information.

    Returns
    -------
    perf : dict
        Dictionary of collected information on CPU and GPU resource usage,
        with the following keys: {
            'cpu': {'util', 'mem_avail', 'mem_total', 'mem_used', 'mem_pct'}
            'gpu': {'util', 'mem_avail', 'mem_total', 'mem_used', 'mem_pct',
                    'alloc', 'alloc_pct', 'max_alloc', 'max_alloc_pct'}
            }

    Notes
    -----
    Usage values include resources used by non-Kilosort processes. This is done
    so that, in the event of an error or crash, we can see if it was due to
    exceeding system resources regardless of whether Kilosort itself was using
    all of them.

    'util' and '_pct' entries are specified as a percentage. Other entries are
    specified in GB.
    
    """

    # CPU stats
    cpu = psutil.cpu_percent()
    memory = psutil.virtual_memory()
    avail = memory.available / 2**30
    total = memory.total / 2**30
    used = total - avail
    pct = memory.percent

    cpu_stats = {
        'util': cpu, 'mem_avail': avail, 'mem_total': total,
        'mem_used': used, 'mem_pct': pct
    }

    # GPU stats
    if torch.cuda.is_available():
        if _NVML_EXISTS:
            gpu = torch.cuda.utilization()
        else:
            gpu = None
        gpu_avail, gpu_total = torch.cuda.mem_get_info()
        gpu_avail /= (2**30)  # convert bytes -> gb
        gpu_total /= (2**30)
        gpu_used = gpu_total - gpu_avail
        gpu_pct = (gpu_used / gpu_total) * 100
        allocated = torch.cuda.memory_allocated() / 2**30
        alloc_pct = (allocated / gpu_total) * 100
        max_alloc = torch.cuda.max_memory_allocated() / 2**30
        max_pct = (max_alloc / gpu_total) * 100

        gpu_stats = {
            'util': gpu, 'mem_avail': gpu_avail, 'mem_total': gpu_total,
            'mem_used': gpu_used, 'mem_pct': gpu_pct, 'alloc': allocated,
            'alloc_pct': alloc_pct, 'max_alloc': max_alloc,
            'max_alloc_pct': max_pct
        }
    else:
        gpu_stats = None

    perf = {'cpu': cpu_stats, 'gpu': gpu_stats}
    return perf


def log_performance(log=None, level=None, header=None, reset=False):
    """Log usage information for cpu, memory, gpu, and gpu memory.

    Parameters
    ----------
    log : logging.Logger; optional.
        Logger object used to write the text. If not provided, the logger for
        `kilosort.utils` will be used.
    level : str; optional.
        Logging level to use. By default, 'debug' will be used. See documentation
        for the built-in `logging` module for information on logging levels.
    header : str; optional.
        Text to output before usage information. For example, an iteration number
        when logging within a loop.
    reset : bool; default=False.
        If True, reset peak cuda memory stats after logging report.

    Notes
    -----
    Usage values include resources used by non-Kilosort processes. This is done
    so that, in the event of an error or crash, we can see if it was due to
    exceeding system resources regardless of whether Kilosort itself was using
    all of them.

    """

    perf = get_performance()

    if log is None:
        log = logger
    if level is None:
        level = 'debug'
    if header is not None:
        getattr(log, level)(' ')
        getattr(log, level)(f'{header}')

    util = perf['cpu']['util']
    avail = perf['cpu']['mem_avail']
    total = perf['cpu']['mem_total']
    used = perf['cpu']['mem_used']
    pct = perf['cpu']['mem_pct']

    getattr(log, level)('*'*56)
    getattr(log, level)(f'CPU usage:    {util:5.2f} %')
    getattr(log, level)(f'Mem used:     {pct:5.2f} %     | {used:10.2f} GB')
    getattr(log, level)(f'Mem avail:    {avail:5.2f} / {total:5.2f} GB')
    getattr(log, level)('-'*54)

    if perf['gpu'] is not None:
        gpu_util = perf['gpu']['util']
        if gpu_util is not None:
            getattr(log, level)(f'GPU usage:    {gpu_util:5.2f} %')
        else:
            getattr(log, level)(f'GPU usage:    `conda install pynvml` for GPU usage')

        gpu_total = perf['gpu']['mem_total']
        gpu_used = perf['gpu']['mem_used']
        gpu_pct = perf['gpu']['mem_pct']
        allocated = perf['gpu']['alloc']
        alloc_pct = perf['gpu']['alloc_pct']
        max_alloc = perf['gpu']['max_alloc']
        max_pct = perf['gpu']['max_alloc_pct']

        getattr(log, level)(f'GPU memory:   {gpu_pct:5.2f} %     |{gpu_used:10.2f}   / {gpu_total:8.2f} GB')
        getattr(log, level)(f'Allocated:    {alloc_pct:5.2f} %     |{allocated:10.2f}   / {gpu_total:8.2f} GB')
        getattr(log, level)(f'Max alloc:    {max_pct:5.2f} %     |{max_alloc:10.2f}   / {gpu_total:8.2f} GB')
    else:
        getattr(log, level)('GPU usage:    N/A')
        getattr(log, level)('GPU memory:   N/A')

    getattr(log, level)('*'*56)

    if reset and torch.cuda.is_available():
        torch.cuda.reset_peak_memory_stats()


def log_cuda_details(log=None):
    """Log a detailed summary of cuda stats from `torch.cuda.memory_summary`."""
    if log is None: log = logger
    if torch.cuda.is_available():
        log.debug(f'\n\n{torch.cuda.memory_summary(abbreviated=True)}\n')


def log_sorting_summary(ops, log=None, level=None):
    """Log a summary of units found, runtime and resource usage after sorting.
    
    Parameters
    ----------
    ops : dict
        Dictionary storing settings and results for all algorithmic steps
        (see `kilosort.run_kilosort`). In particular, `ops` as returned by
        `run_kilosort` after sorting is finished, or loaded from an existing
        result from v4.0.38 or later.
    log : logging.Logger; optional.
        Logger object used to write the text. If not provided, the logger for
        `kilosort.utils` will be used.
    level : str; optional.
        Logging level to use. By default, 'debug' will be used. See documentation
        for the built-in `logging` module for information on logging levels.
    
    Notes
    -----
    Usage values include resources used by non-Kilosort processes. This is done
    so that, in the event of an error or crash, we can see if it was due to
    exceeding system resources regardless of whether Kilosort itself was using
    all of them.
        
    """

    if ops.get('runtime_postproc', None) is None:
        raise ValueError('Ops must contain runtime stats (v4.0.38 or later).')

    if log is None:
        log = logger
    if level is None:
        level = 'debug'
    log_fn = getattr(log, level)

    step_keys = {
        'preprocessing': 'preproc', 'drift corr': 'drift',
        'spike det. (univ)': 'st0', 'cluster (temp)': 'clu0',
        'spike det. (learn)': 'st', 'cluster (final)': 'clu',
        'cluster merge': 'merge', 'postprocessing': 'postproc'
    }

    log_fn(' ')
    log_fn('*'*56)
    log_fn('Sorting summary')
    log_fn('-'*56)
    log_fn(f"{'Total number of units:':<30}{ops['n_units_total']:>25}")
    log_fn(f"{'Number of good units:':<30}{ops['n_units_good']:>25}")
    log_fn(f"{'Number of spikes':<30}{ops['n_spikes']:>25}")
    log_fn(f"{'Mean abs. drift':<30}{ops['mean_drift']:>22.1f} um")
    log_fn(f"{'Total runtime':<30}{ops['runtime']:>24.2f}s")

    log_fn(' ')
    log_fn('Runtime by step')
    log_fn('-'*56)
    for name, key in step_keys.items():
        log_fn(_format_runtime(ops, name, key))

    gpu = True if ops['usage_postproc']['gpu'] is not None else False
    log_fn(' ')
    log_fn('Memory usage by step')
    log_fn('-'*56)
    for name, key in step_keys.items():
        log_fn(_format_usage(ops, name, key, gpu))
    log_fn('*'*56)

def _format_runtime(ops, name, key):
    total = ops['runtime']
    step = ops['runtime_' + key]
    pct = step / total * 100
    return f"{(name + ':'):<27} {step:>15.1f}s {('(' + f'{pct:.2f}' + ') %'):>10}"

def _format_usage(ops, name, key, gpu):
    s = (f"{(name + ':'):<23}"
         + f"sys  {ops['usage_' + key]['cpu']['mem_used']:>6.1f} GB  |  ")
    if gpu:
        s += f"gpu  {ops['usage_' + key]['gpu']['max_alloc']:>5.1f} GB"
    else:
        s += f"gpu  {'N/A':>7}   "
    return s


def log_thread_count(log=None):
    if log is None:
        log = logger

    if sys.platform == 'linux':
        result = subprocess.run(['ps', '-o', 'thcount', f'{os.getpid()}'], stdout=subprocess.PIPE)
        thread_count = result.stdout.decode('utf-8').replace('\n','').replace(' ','').strip('THCNT')
    else:
        thread_count = 'N/A (linux only)'

    log.debug('')
    log.debug('--------------------')
    log.debug(f'Process thread count: {thread_count}')
    log.debug(f'num torch threads: {torch.get_num_threads()}')
    log.debug('--------------------\n')


def probe_as_string(probe):
    """Format probe dictionary as copy-pasteable-to-code string."""

    # Set numpy to print full arrays
    opt = np.get_printoptions()
    np.set_printoptions(threshold=np.inf)
    
    p = pprint.pformat(probe, indent=4, sort_dicts=False)
    # insert `np.` so that text can be copied directly to code
    p = 'np.array'.join(p.split('array'))
    p = 'dtype=np.'.join(p.split('dtype='))
    probe_text = "probe = "
    # Put curly braces on separate lines
    probe_text += p[0] + '\n ' + p[1:-1] + '\n' + p[-1]

    # Revert numpy settings
    np.set_printoptions(**opt)

    return probe_text


def ops_as_string(ops):
    """Format ops dictionary as copy-pasteable-to-code string.
    
    Notes
    -----
    Keys for `settings` and `probe` are removed since they contain a lot of
    redundant information and are difficult to format in a nested way. See
    `probe_as_string` for printing probe information.
    
    """

    ops_copy = ops.copy()
    probe_keys = list(ops['probe'].keys())
    for k in ['settings', 'probe'] + probe_keys:
        _ = ops_copy.pop(k)
    n_files = len(ops_copy['filename'])
    if n_files > 5:
        ops_copy['filename'] = ops_copy['filename'][:5] + [f'... ({n_files} total files)']

    ops_text = "ops = "
    p = pprint.pformat(ops_copy, indent=4, sort_dicts=False)
    # Put curly braces on separate lines
    ops_text += p[0] + '\n ' + p[1:-1] + '\n' + p[-1]

    return ops_text
